#ifndef AGENT_HH
#define AGENT_HH

//----------------------------------------------------------------------------
// pacman-like game
// copyright 2010 christopher rasmussen
// university of delaware
// v. 1.06
// * attempt_move() added
// * last_position added 
// * initial_position, default_draw_color added
// * blue_ghost_wait_counter added to GameState
// * added object_last_position to GameState
// * split GameState off into separate .hh
// * added execute_move(); attempt_move() no longer makes move

//----------------------------------------------------------------------------

#include "BlockMaze.hh"
#include "Utils.hh"
#include "GameState.hh"

//----------------------------------------------------------------------------
//----------------------------------------------------------------------------

#define UTILITY_RUNNER_WIN    100
#define UTILITY_GHOST_WIN    -100

const char NORTH_KEY     = 'w';
const char SOUTH_KEY     = 's';
const char WEST_KEY      = 'a';
const char EAST_KEY      = 'd';

const char DRAW_DISTANCE_KEY  = 'o';
const char DRAW_PATH_KEY      = 'p';
const char DRAW_NAME_KEY      = 'n';

//----------------------------------------------------------------------------

class GameState;

void BFS(Point_2D, BlockMaze *, vector <Point_2D> * = NULL);
void BFS_expand(Point_2D, Point_2D &, queue <Point_2D> &, BlockMaze *, vector <Point_2D> *);
bool compute_BFS_path(Point_2D, vector <Point_2D> &, BlockMaze *);

// these assume BFS has been called for start position already

bool closest_power_pill(int &, int &, const GameState &, BlockMaze *);
bool closest_ghost(int &, int &, const GameState &, BlockMaze *);
bool closest_runner(int &, int &, const GameState &, BlockMaze *);

// name-inference functions

bool is_ghost(string);
bool is_runner(string);
bool is_power_pill(string);

// ARENA

bool closest_available_even_power_pill(vector <bool> &, int &, int &, const GameState &, BlockMaze *);
bool closest_available_odd_power_pill(vector <bool> &, int &, int &, const GameState &, BlockMaze *);
bool closest_even_power_pill(int &, int &, const GameState &, BlockMaze *);
bool closest_odd_power_pill(int &, int &, const GameState &, BlockMaze *);
bool closest_available_even_runner(vector <bool> &, int &, int &, const GameState &, BlockMaze *);
bool closest_available_odd_runner(vector <bool> &, int &, int &, const GameState &, BlockMaze *);
bool closest_even_runner(int &, int &, const GameState &, BlockMaze *);
bool closest_odd_runner(int &, int &, const GameState &, BlockMaze *);

bool is_odd_team(string);
bool is_even_team(string);

//----------------------------------------------------------------------------
//----------------------------------------------------------------------------

class Agent
{
public:

  Agent(string, char, int, int = 0, int = 0);
  virtual void update(const GameState &, BlockMaze *) = 0;
  void draw(BlockMaze *);
  void draw_debug(BlockMaze *);
  void handle_key(char, BlockMaze *);
  bool attempt_move(int, int, BlockMaze *);
  void execute_move(BlockMaze *);
  void ghost_scatter_plan(const GameState &, BlockMaze *);  // smart ghosts use this to get away during "blue time"
  void reset();

  // key identifiers

  string name;
  Point_2D position, last_position, initial_position;

  bool move_approved;          // is there a move stored here?
  Point_2D move_position;      // move stored here (null move if not approved by attempt_move())

  int terrain_wait_counter;

  // display variables

  char draw_glyph;
  int draw_color, default_draw_color;

  bool do_draw_distances;
  bool do_draw_path;
  bool do_draw_name;

  // optional "AI" variables and data structures

  vector <Point_2D> path;
  int turn_number;

};

//----------------------------------------------------------------------------
//----------------------------------------------------------------------------

#endif
