//----------------------------------------------------------------------------
// pacman-like game
// copyright 2010 christopher rasmussen
// university of delaware
// v. 1.0

//----------------------------------------------------------------------------

#include "ArenaAgent.hh"

//----------------------------------------------------------------------------
//----------------------------------------------------------------------------

//#define ARENA_PRINT_DEBUG

// STUDENTS SET THIS

const int ARENA_MAXIMUM_DEPTH_FOR_MINIMAX = 1;

//----------------------------------------------------------------------------
//----------------------------------------------------------------------------

// upper bound on number of moves required to eat n power pills

int arena_max_path_length(int n, BlockMaze *M)
{ 
  return n * M->num_empty_spaces;
}

//----------------------------------------------------------------------------

float max_utility(const GameState & GS, BlockMaze *M)
{
  return arena_max_path_length(GS.initial_num_power_pills / 2, M);
}

//----------------------------------------------------------------------------

// is this the end of the "game"?  if so, who won?

// EVEN -> MAX
// ODD  -> MIN

// assuming even division of power pills between odd and even side

bool ArenaAgent::terminal_state(float & utility, const GameState & GS, BlockMaze *M)
{
  if (GS.arena_check_for_even_level_cleared(M)) {
    utility = max_utility(GS, M);
    return true;
  }
  else if (GS.arena_check_for_odd_level_cleared(M)) {
    utility = -max_utility(GS, M);
    return true;
  }
  else {
    return false;
  }
}

//----------------------------------------------------------------------------

// should we stop searching and apply eval() to this GameState?

bool ArenaAgent::cutoff_test(int depth, const GameState & GS, BlockMaze *M)
{
  if (depth >= ARENA_MAXIMUM_DEPTH_FOR_MINIMAX) 
    return true;
  else 
    return false;
}

//----------------------------------------------------------------------------

class PathCompare
{
public:
  bool operator() (vector <Point_2D> & lhs, vector <Point_2D> & rhs) const
  {
    return lhs.size() > rhs.size();
  }
};

//----------------------------------------------------------------------------

// look for closest pill with no opponent on path to it

// assuming BFS has been called from start position already WITH EVEN OBSTACLES

bool smart_select_available_even_power_pill(vector <bool> & taken, int & selected_pill_distance, int & selected_pill_index, const GameState & GS, BlockMaze *M)
{
  int i, j, k;
  vector <Point_2D> P;
  Point_2D p;
  priority_queue <vector <Point_2D>, vector < vector <Point_2D> >, PathCompare> PQ;      
  
  for (i = 0; i < GS.object_name.size(); i++) 
    if (is_power_pill(GS.object_name[i]) && M->is_odd_side(GS.object_position[i]) && !taken[i] &&
	compute_BFS_path(GS.object_position[i], P, M)) {
      p.row = -1;
      p.col = i;
      P.push_back(p);
      PQ.push(P);
    }

  if (!PQ.empty()) {

    P = PQ.top(); PQ.pop();
    p = P.back();  P.pop_back();

    selected_pill_index = p.col;
    selected_pill_distance = P.size();
    return true;
  }

  return false;
}

//----------------------------------------------------------------------------

// look for closest pill with no opponent on path to it

// assuming BFS has been called from start position already WITH ODD OBSTACLES

bool smart_select_available_odd_power_pill(vector <bool> & taken, int & selected_pill_distance, int & selected_pill_index, const GameState & GS, BlockMaze *M)
{
  int i, j, k;
  vector <Point_2D> P;
  Point_2D p;
  priority_queue <vector <Point_2D>, vector < vector <Point_2D> >, PathCompare> PQ;      
  
  for (i = 0; i < GS.object_name.size(); i++) 
    if (is_power_pill(GS.object_name[i]) && M->is_even_side(GS.object_position[i]) && !taken[i] &&
	compute_BFS_path(GS.object_position[i], P, M)) {
      p.row = -1;
      p.col = i;
      P.push_back(p);
      PQ.push(P);
    }

  if (!PQ.empty()) {

    P = PQ.top(); PQ.pop();
    p = P.back();  P.pop_back();

    selected_pill_index = p.col;
    selected_pill_distance = P.size();
    return true;
  }

  return false;
}

//----------------------------------------------------------------------------

// STUDENTS WRITE THIS FOR PROJECT 3

#define FIGHT_OR_FLIGHT_DISTANCE    5   // how close is too close for opposing team member
//#define FIGHT_OR_FLIGHT_DISTANCE    0   // how close is too close for opposing team member

// how good is this GameState for the EVEN (MAX) team and the ODD (MIN) team?

float ArenaAgent::eval(const GameState & GS, BlockMaze *M)
{
  int i;
  int closest_pill_distance, closest_pill_index;
  int selected_pill_distance, selected_pill_index;
  int closest_runner_distance, closest_runner_index;
  float utility;
  vector <bool> taken;
  vector <Point_2D> odd_obstacles;
  vector <Point_2D> even_obstacles;

  taken.resize(GS.object_name.size());
  for (i = 0; i < GS.object_name.size(); i++) 
    taken[i] = false;

  odd_obstacles.clear();
  for (i = 0; i < GS.object_name.size(); i++) 
    if (is_even_team(GS.object_name[i]))
      odd_obstacles.push_back(GS.object_position[i]);

  even_obstacles.clear();
  for (i = 0; i < GS.object_name.size(); i++) 
    if (is_odd_team(GS.object_name[i]))
      even_obstacles.push_back(GS.object_position[i]);
  
  utility = 0;

  // collect power pills, avoid others when on offense, attack others on defense

  for (i = 0; i < GS.object_name.size(); i++) {

    // EVEN = MAX

    if (is_even_team(GS.object_name[i])) {

      BFS(GS.object_position[i], M);

      // fight or flight!

      if (closest_odd_runner(closest_runner_distance, closest_runner_index, GS, M) &&
	  closest_runner_distance <= FIGHT_OR_FLIGHT_DISTANCE) {
	if (M->is_even_side(GS.object_position[closest_runner_index])) {
	  utility -= closest_runner_distance;  // home turf -> attack: closer is better for EVEN
	}
	else {
	  utility += closest_runner_distance;  // away -> flee: farther is better for EVEN
	}
      }

      // go for power pill!

      else {

	BFS(GS.object_position[i], M, &even_obstacles);

	if (smart_select_available_even_power_pill(taken, selected_pill_distance, selected_pill_index, GS, M)) {
	  taken[selected_pill_index] = true;
	  utility -= selected_pill_distance;        // the farther, the better for ODD (MIN)
	}

	// flummoxed!

	else {

	}
      }
    }

    // ODD = MIN

    else if (is_odd_team(GS.object_name[i])) {

      BFS(GS.object_position[i], M);

      // fight or flight!

      if (closest_even_runner(closest_runner_distance, closest_runner_index, GS, M) &&
	  closest_runner_distance <= FIGHT_OR_FLIGHT_DISTANCE) {
	if (M->is_odd_side(GS.object_position[closest_runner_index])) {
	  utility += closest_runner_distance;  // home turf -> attack: closer is better for EVEN
	}
	else {
	  utility -= closest_runner_distance;  // away -> flee: farther is better for EVEN
	}
      }

      // go for power pill!

      else {

	BFS(GS.object_position[i], M, &odd_obstacles);

	if (smart_select_available_odd_power_pill(taken, selected_pill_distance, selected_pill_index, GS, M)) {
	  taken[selected_pill_index] = true;
	  utility += selected_pill_distance;      // the farther, the better for ODD (MIN)
	}
	
	// flummoxed!
	
	else {

	}
      }
    }
  }

  // incorporate all untaken pills into utility

  int pill_penalty = arena_max_path_length(1, M);  // farthest any runner could be from this pill

  for (i = 0; i < GS.object_name.size(); i++) {

    if (is_power_pill(GS.object_name[i]) && !taken[i]) {

      // EVEN pill untaken -- bad for EVEN

      if (M->is_odd_side(GS.object_position[i])) {
	utility -= pill_penalty;  
      }

      // ODD pill untaken -- bad for ODD

      else {
	utility += pill_penalty;  
      }
    }
  }

  return utility;
}

//----------------------------------------------------------------------------

// OPTIONAL HELPER FUNCTION THAT I WROTE

void ArenaAgent::update_best_minimax_move(bool even_turn, int move_depth, int move_agent_index, int move_delta_row, int move_delta_col, const GameState & GS, BlockMaze *M, float & best_value, Point_2D & best_move)
{
  Point_2D move_pos, temp_pos;
  GameState move_GS;
  float move_value;
  int which_pill, which_runner;
  int cur_delta_row, cur_delta_col;

  cur_delta_row = GS.object_position[move_agent_index].row - GS.object_last_position[move_agent_index].row;
  cur_delta_col = GS.object_position[move_agent_index].col - GS.object_last_position[move_agent_index].col;

  // no reverses allowed unless dead end!

  if ((fabs(cur_delta_row - move_delta_row) > 1 || fabs(cur_delta_col - move_delta_col) > 1))
    return;

  move_pos.row = GS.object_position[move_agent_index].row + move_delta_row;
  move_pos.col = GS.object_position[move_agent_index].col + move_delta_col;

  if (!M->is_wall(move_pos.row, move_pos.col)) {

#ifdef ARENA_PRINT_DEBUG
    if (even_turn)
      printf(">>> EVEN successor (%i, %i)\n", move_pos.row, move_pos.col);
    else
      printf(">>> ODD successor (%i, %i)\n", move_pos.row, move_pos.col);
#endif

    move_GS = GS;
    move_GS.object_last_position[move_agent_index] = move_GS.object_position[move_agent_index];
    move_GS.object_position[move_agent_index] = move_pos;

    if (move_GS.arena_check_for_power_pill_eaten(which_pill, which_runner, M)) 
      move_GS.remove_object(which_pill);

    // what is minimax value of this "successor" state and
    // is it better than what we've seen so far?

    minimax_value(!even_turn, move_depth + 1, move_GS, M, move_value, temp_pos);

    // if it's EVEN's turn, best is MAX

    if (even_turn) {
      if (move_value > best_value || (move_value == best_value && coin_flip())) {
	best_value = move_value;
	best_move = move_pos;
      }
    }

    // if it's ODD's turn, best is MIN

    else {
      if (move_value < best_value || (move_value == best_value && coin_flip())) {
	best_value = move_value;
	best_move = move_pos;
      }
    }
  }
}

//----------------------------------------------------------------------------

// STUDENTS WRITE THIS FOR PROJECT 3

// runner_turn: is this level of game tree EVEN (MAX) making a move or ODD (MIN)?
// depth: how far down the game tree are we (0 = root)
// GS, M: state of game to compute value for
// best_value: what is the computed value? (should be filled in by this function)
// best_move: where should the agent whose turn it is go? (should be filled in by this function)

// return value = true: best_move has been filled in (i.e., there is a move to make)
// return value = false: there is no best_move because GS is a leaf of game tree
 
bool ArenaAgent::minimax_value(bool even_turn, int depth, const GameState & GS, BlockMaze *M, float & best_value, Point_2D & best_move)
{
  int i, index, this_runner_index, closest_opposing_runner_distance, closest_opposing_runner_index;
  bool is_opposing_runner;

#ifdef ARENA_PRINT_DEBUG
  printf("in MINIMAX depth = %i, even = %i\n", depth, even_turn);
#endif

  // which object is ME?
  
  for (i = 0; i < GS.object_name.size(); i++) 
    if (GS.object_name[i] == name) {
      this_runner_index = i;
      break;
    }

  // is the game over?
  
  if (terminal_state(best_value, GS, M))   // if true, utility is filled in
    return false;

  // are we at the depth limit of our search?

  else if (cutoff_test(depth, GS, M)) {
    best_value = eval(GS, M);
#ifdef ARENA_PRINT_DEBUG
    printf("EVAL = %.1f (%i, %i)\n", best_value, best_move.row, best_move.col);
#endif
    return false;
  }

  else {

    // EVEN's turn: return MAX over s in successors(GS) of minimax_value(s)
    
    // turn-taker is EVEN ME

    if (even_turn) {

      if (is_even_team(name)) {
	index = this_runner_index;
#ifdef ARENA_PRINT_DEBUG
	printf("EVEN TURN, EVEN AGENT\n");
#endif
      }

      // I am ODD, so turn-taker is nearest agent on opposing EVEN team

      else {
#ifdef ARENA_PRINT_DEBUG
	printf("EVEN TURN, ODD AGENT\n");
#endif
	BFS(GS.object_position[this_runner_index], M);
	if (!closest_even_runner(closest_opposing_runner_distance, index, GS, M))
	  return minimax_value(!even_turn, depth + 1, GS, M, best_value, best_move);
      }

      best_value = -(max_utility(GS, M) + 1);
#ifdef ARENA_PRINT_DEBUG
      printf("MAX SETTING BEST = %.1f\n", best_value);
#endif
    }

    // ODD's turn: return MIN over s in successors(GS) of minimax_value(s)

    // turn-taker is ODD ME

    else {

      if (is_odd_team(name)) {
	index = this_runner_index;
#ifdef ARENA_PRINT_DEBUG
	printf("ODD TURN, ODD AGENT\n");
#endif
      }

      // I am EVEN, so turn-taker is nearest agent on opposing ODD team

      else {
#ifdef ARENA_PRINT_DEBUG
	printf("ODD TURN, EVEN AGENT\n");
#endif
	BFS(GS.object_position[this_runner_index], M);
	if (!closest_odd_runner(closest_opposing_runner_distance, index, GS, M))
	  return minimax_value(!even_turn, depth + 1, GS, M, best_value, best_move);
      }

      best_value = max_utility(GS, M) + 1;
#ifdef ARENA_PRINT_DEBUG
      printf("MIN SETTING BEST = %.1f\n", best_value);
#endif
    }

    update_best_minimax_move(even_turn, depth, index, -1,  0, GS, M, best_value, best_move);
    update_best_minimax_move(even_turn, depth, index, +1,  0, GS, M, best_value, best_move);
    update_best_minimax_move(even_turn, depth, index,  0, +1, GS, M, best_value, best_move);
    update_best_minimax_move(even_turn, depth, index,  0, -1, GS, M, best_value, best_move);
    
    return true;
  }
}

//----------------------------------------------------------------------------

// ALL STUDENTS WRITE THIS FOR PROJECT 3

// ad-hoc ok for "Ghosts" scenario, but must use minimax for "Power Pills Work!" and "The Arena"

void ArenaAgent::update(const GameState & GS, BlockMaze *M)
{
  // uncomment for total manual control
  //  return;

  // uncomment to use minimax method

  if (minimax_value(is_even_team(name), 0, GS, M, best_minimax_value, best_minimax_move)) {
#ifdef ARENA_PRINT_DEBUG
    printf("MOVING value %.1f to (%i, %i) ---------------------->\n", best_minimax_value, best_minimax_move.row, best_minimax_move.col);
#endif
    attempt_move(best_minimax_move.row, best_minimax_move.col, M);
  }

}

//----------------------------------------------------------------------------
//----------------------------------------------------------------------------

