//----------------------------------------------------------------------------
// union-find framework for disjoint sets
// copyright 2010 christopher rasmussen
// university of delaware
// v. 1.0
//----------------------------------------------------------------------------

#include "DisjointSet.hh"

//----------------------------------------------------------------------------
//----------------------------------------------------------------------------

DisjointSet::DisjointSet(int num_elements)
{
  S.clear();
  S.resize(num_elements);

  for (int i = 0; i < S.size(); i++)
    S[i] = -1;

  num_equivalence_classes = S.size();
}

//----------------------------------------------------------------------------

void DisjointSet::print()
{
  cout << num_equivalence_classes << " equivalence classes\n";

  for (int i = 0; i < S.size(); i++) {
    if (S[i] < 0)
      cout << "* ";
    else
      cout << "  ";
    cout << i << ": " << S[i] << endl;
  }
}

//----------------------------------------------------------------------------

// "dumb" union

void DisjointSet::merge(int x, int y)
{
  int root_x = find(x);
  int root_y = find(y);

  if (root_x != root_y) {
    S[root_y] = root_x;
    num_equivalence_classes--;
  }
}

//----------------------------------------------------------------------------

// includes path correction

int DisjointSet::find(int x)
{
  if (x < 0 || x >= S.size()) {
    cout << "find(): " << x << " is not a set element\n";
    exit(1);
  }

  if (S[x] < 0) 
    return x;
  else
    return (S[x] = find(S[x]));
}

//----------------------------------------------------------------------------
//----------------------------------------------------------------------------
