//----------------------------------------------------------------------------
// pacman-like game
// copyright 2010 christopher rasmussen
// university of delaware
// v. 1.01
// * all check_*() functions in GameState

//----------------------------------------------------------------------------

#include "GameState.hh"

//----------------------------------------------------------------------------
//----------------------------------------------------------------------------

// "compress" information about variable parts of game state--where agents are, where
// remaining power pills are, etc., into single class object

void GameState::update(vector <Agent *> & agent_list, vector <Point_2D> & power_pill_list)
{
  int i;

  object_position.clear();
  object_last_position.clear();
  object_name.clear();

  for (i = 0; i < agent_list.size(); i++) {
    object_position.push_back(agent_list[i]->position);
    object_last_position.push_back(agent_list[i]->last_position);
    object_name.push_back(agent_list[i]->name);
  }

  for (i = 0; i < power_pill_list.size(); i++) {
    object_position.push_back(power_pill_list[i]);
    object_last_position.push_back(power_pill_list[i]);
    object_name.push_back("power pill");
  }
}

//----------------------------------------------------------------------------

void GameState::print() const
{
  int i;

  for (i = 0; i < object_name.size(); i++) {
    cout << i << ": " << object_name[i] << " (" << object_position[i].row << ", " << object_position[i].col << ")\n";
//     object_position.push_back(agent_list[i]->position);
//     object_last_position.push_back(agent_list[i]->last_position);
//     object_name.push_back(agent_list[i]->name);
  }
}

//----------------------------------------------------------------------------

// take an object out of the game (be it an agent or a power pill)

void GameState::remove_object(int i)
{
  object_name.erase(object_name.begin() + i);
  object_position.erase(object_position.begin() + i);
  object_last_position.erase(object_last_position.begin() + i);
}

//----------------------------------------------------------------------------

// how many power pills are left?

int GameState::num_power_pills() const
{
  int i, n;

  for (i = n = 0; i < object_name.size(); i++)
    if (is_power_pill(object_name[i]))
      n++;

  return n;
}

//----------------------------------------------------------------------------

// has a runner just eaten any power pill?

bool GameState::check_for_power_pill_eaten(int & which_pill, int & which_runner) const
{
  int i, j;

  for (i = 0; i < object_name.size(); i++) {
    if (is_power_pill(object_name[i])) {
      for (j = 0; j < object_name.size(); j++) {
	if (is_runner(object_name[j]) && object_position[i] == object_position[j]) {
	  which_pill = i;
	  which_runner = j;
	  return true;
	}
      }
    }
  }

  return false;
}

//----------------------------------------------------------------------------

// did a runner consume a power pill on the correct side of the arena?

bool GameState::arena_check_for_power_pill_eaten(int & which_pill, int & which_runner, BlockMaze *M) const
{
  int i, j;

  for (i = 0; i < object_name.size(); i++) {
    if (is_power_pill(object_name[i])) {
      for (j = 0; j < object_name.size(); j++) {
	if (object_position[j] == object_position[i] &&
	  ((is_odd_team(object_name[j]) && M->is_even_side(object_position[i])) ||
	   (is_even_team(object_name[j]) && M->is_odd_side(object_position[i])))) {
	  which_pill = i;
	  which_runner = j;
	  return true;
	}
      }
    }
  }

  return false;
}

//----------------------------------------------------------------------------

// is any ghost agent position coincident with runner agent position?
// or did they switch positions (aka crossover)?  if so, they must have collided...
// if there was/is a collision, return index of ghost involved

bool GameState::check_for_runner_caught(int & which_ghost) const
{
  int i;
  Point_2D runner_position, runner_last_position;

  for (i = 0; i < object_name.size(); i++) 
    if (is_runner(object_name[i])) {
      runner_position = object_position[i];
      runner_last_position = object_last_position[i];
      break;
    }

  for (i = 0; i < object_name.size(); i++) 
    if (is_ghost(object_name[i]) &&
	((object_position[i] == runner_position) ||            // same place now
	 (object_position[i] == runner_last_position && 
	  object_last_position[i] == runner_position))) {  // crossover occurred
      which_ghost = i;
      return true;
    }

  return false;
}

//----------------------------------------------------------------------------

// is any agent position "on the blue side" coincident with "defending" agent position?
// or did they switch positions (aka crossover)?  if so, they must have collided...
// if there was/is a collision, return index of agent involved so it can be "sent home"

// this only detects one agent being caught--must be called multiple times in case
// there are simultaneous captures

bool GameState::arena_check_for_runner_caught(int & which_caught, BlockMaze *M) const
{
  int i, j;

  for (i = 0; i < object_name.size(); i++)
    for (j = 0; j < object_name.size(); j++) 
      if ((object_position[i] == object_position[j] || 
	   (object_position[i] == object_last_position[j] && object_last_position[i] == object_position[j])) &&
	  ((is_odd_team(object_name[i]) && is_even_team(object_name[j]) && M->is_even_side(object_position[i])) ||
	   (is_even_team(object_name[i]) && is_odd_team(object_name[j]) && M->is_odd_side(object_position[i])))) {
	which_caught = i;
	return true;
      }

  return false;
}

//----------------------------------------------------------------------------

bool GameState::arena_check_for_odd_level_cleared(BlockMaze *M) const
{
  for (int i = 0; i < object_name.size(); i++)
    if (is_power_pill(object_name[i]) && M->is_even_side(object_position[i]))
      return false;

  return true;
}

//----------------------------------------------------------------------------

bool GameState::arena_check_for_even_level_cleared(BlockMaze *M) const
{
  for (int i = 0; i < object_name.size(); i++)
    if (is_power_pill(object_name[i]) && M->is_odd_side(object_position[i]))
      return false;

  return true;
}

//----------------------------------------------------------------------------
//----------------------------------------------------------------------------

// are all power pills gone?

bool GameState::check_for_level_cleared() const
{
  for (int i = 0; i < object_name.size(); i++)
    if (is_power_pill(object_name[i]))
      return false;

  return true;
}

//----------------------------------------------------------------------------

// is an object whose name contains string s located at (row, col)?

bool GameState::object_here(string s, int row, int col) const
{
  int i;

  for (i = 0; i < object_name.size(); i++)
    if (object_name[i].find(s) != string::npos &&
	object_position[i].row == row && object_position[i].col == col)
      return true;

  return false;
} 

//----------------------------------------------------------------------------
//----------------------------------------------------------------------------

