//----------------------------------------------------------------------------
// pacman-like game
// copyright 2010 christopher rasmussen
// university of delaware
// v 1.0
//----------------------------------------------------------------------------

#include "GhostAgent.hh"

//----------------------------------------------------------------------------
//----------------------------------------------------------------------------

// random destinations only, 50% speed

void ClydeGhostAgent::update(const GameState & GS, BlockMaze *M)
{
  // "rest" every 2th turn => 50% speed

  turn_number++;

  if (!(turn_number % 2))
    return;

  // pick random target

  // we have no current plan, so make one

  if (path.size() == 0) {

    Point_2D p;
    M->pick_random_empty_space(p.row, p.col);

    // label all vertices with min-distance path length

    BFS(position, M);

    // compute plan to random spot

    compute_BFS_path(p, path, M);
  }

  // we have a plan...move!

  // get rid of null move if any

  if (path[0].x == position.x && path[0].y == position.y)
    path.erase(path.begin());    

  // try to execute next step in plan...

  if (path.size() && attempt_move(path[0].y, path[0].x, M))
    path.erase(path.begin());    // erase that move from plan if successful

}

//----------------------------------------------------------------------------

// how get away from the runner when it's "blue time"
// does not actually move--just makes a plan

void Agent::ghost_scatter_plan(const GameState & GS, BlockMaze *M)
{
  int i, row, col, max_dist, ghost_dist;
  Point_2D max_dist_p;

  path.clear();

  for (i = 0; i < GS.object_position.size(); i++)
    if (is_runner(GS.object_name[i])) {
      
      // BFS from runner point of view
      
      BFS(GS.object_position[i], M);
      
      // save most distant point
      
      ghost_dist = M->vertex[position.row][position.col].distance;
      max_dist = 0;
      for (row = 0; row < M->num_rows; row++)
	for (col = 0; col < M->num_cols; col++)
	  if (!M->vertex[row][col].is_wall() && M->vertex[row][col].distance > max_dist) {
	    max_dist = M->vertex[row][col].distance;
	      max_dist_p.row = row;
	      max_dist_p.col = col;
	  }


      // BFS from ghost point of view
      
      BFS(position, M);
      
      // plan path to most distant point 
      
      compute_BFS_path(max_dist_p, path, M);

      break;
    }
}

//----------------------------------------------------------------------------

// 1/2 clyde, 1/2 cut off runner on way to nearest power pill, 50% speed

void InkyGhostAgent::update(const GameState & GS, BlockMaze *M)
{
  // "rest" every other turn => 50% speed

  turn_number++;

  if (!(turn_number % 2))
    return;

  // changes modes?

  if (mode_wait_counter == 0) {
    inky_mode = rand() % 2;
    mode_wait_counter = rand_in_range(10, 20);  // be this way for 10-20 turns
  }

  // inky mode: cut off the runner

  if (inky_mode) {

    int dist;
    int min_dist = 100000;
    int min_dist_index, path_target_index;
    vector <Point_2D> runner_path;
    bool is_power_pill; 
    int closest_pill_distance, closest_pill_index;
    
    for (int i = 0; i < GS.object_position.size(); i++)
      if (is_runner(GS.object_name[i])) {
	
	BFS(GS.object_position[i], M);

	is_power_pill = closest_power_pill(closest_pill_distance, closest_pill_index, GS, M);

	if (is_power_pill)
	  compute_BFS_path(GS.object_position[closest_pill_index], runner_path, M);
	else
	  return;

	BFS(position, M);
	path_target_index = rand_in_range(runner_path.size() / 4, 3 * runner_path.size() / 4);
	compute_BFS_path(runner_path[path_target_index], path, M);
	
	break;
      }
  }

  // clyde mode: move randomly

  else {

    if (path.size() == 0) {

      Point_2D p;
      M->pick_random_empty_space(p.row, p.col);
      BFS(position, M);
      compute_BFS_path(p, path, M);
    }
  }
  
  mode_wait_counter--;

  // get rid of null move if any

  if (path[0].x == position.x && path[0].y == position.y)
    path.erase(path.begin());    
  
  if (path.size() && attempt_move(path[0].y, path[0].x, M))
    path.erase(path.begin());    // erase that move from plan if successful
}

//----------------------------------------------------------------------------

// always go straight for the runner, 50% speed

void BlinkyGhostAgent::update(const GameState & GS, BlockMaze *M)
{
  // "rest" every other turn => 50% speed

  turn_number++;

  if (!(turn_number % 2))
    return;

  // "blue time"?

  if (GS.blue_ghost_wait_counter > 0) 
    ghost_scatter_plan(GS, M);

  else {

    // label all vertices with min-distance path length

    BFS(position, M);

    // compute plan to runner

    for (int i = 0; i < GS.object_position.size(); i++)
      if (is_runner(GS.object_name[i])) {
	compute_BFS_path(GS.object_position[i], path, M);
	break;
      }
  }

  // get rid of null move if any
  
  if (path[0].x == position.x && path[0].y == position.y)
    path.erase(path.begin());    

  // try to execute next step in plan...

  attempt_move(path[0].y, path[0].x, M);
}

//----------------------------------------------------------------------------
//----------------------------------------------------------------------------
