//----------------------------------------------------------------------------
// pacman-like game
// copyright 2010 christopher rasmussen
// university of delaware
// v. 1.0

//----------------------------------------------------------------------------

#include "RunnerAgent.hh"

//----------------------------------------------------------------------------
//----------------------------------------------------------------------------

//#define PRINT_DEBUG

// STUDENTS SET THIS

const int MAXIMUM_DEPTH_FOR_MINIMAX = 2;

//----------------------------------------------------------------------------
//----------------------------------------------------------------------------

// HONORS STUDENTS WRITE THIS FOR LAB 7

void Dijkstra(Point_2D start, BlockMaze *M)
{

}

//----------------------------------------------------------------------------
//----------------------------------------------------------------------------

// upper bound on number of moves required to eat n power pills

int max_path_length(int n, BlockMaze *M)
{ 
  return n * M->num_empty_spaces;
}

//----------------------------------------------------------------------------

// is this the end of the "game"?  if so, who won?

bool RunnerAgent::terminal_state(float & utility, const GameState & GS, BlockMaze *M)
{
  int which_ghost;

#ifdef PRINT_DEBUG
  printf("terminal test\n");
  GS.print();
#endif

  if (GS.check_for_runner_caught(which_ghost) && GS.blue_ghost_wait_counter == 0) {
    //    utility = UTILITY_GHOST_WIN;
    utility = -max_path_length(GS.initial_num_power_pills, M);
#ifdef PRINT_DEBUG
    printf("GHOST WIN\n");
#endif
    return true;
  }
  else if (GS.check_for_level_cleared()) {
    //    utility = UTILITY_RUNNER_WIN; 
    utility = max_path_length(GS.initial_num_power_pills, M);
#ifdef PRINT_DEBUG
    printf("RUNNER WIN\n");
#endif
    return true;
  }
  else {
#ifdef PRINT_DEBUG
    printf("NON-TERMINAL\n");
#endif
    return false;
  }
}

//----------------------------------------------------------------------------

// should we stop searching and apply eval() to this GameState?

bool RunnerAgent::cutoff_test(int depth, const GameState & GS, BlockMaze *M)
{
#ifdef PRINT_DEBUG
  printf("cutoff %i...", depth);
#endif

  if (depth >= MAXIMUM_DEPTH_FOR_MINIMAX) {
#ifdef PRINT_DEBUG
    printf("TRUE\n");
#endif
    //GS.print();
    return true;
  }
  else {
#ifdef PRINT_DEBUG
    printf("FALSE\n");
#endif
    //GS.print();
    return false;
  }
}

//----------------------------------------------------------------------------

// STUDENTS WRITE THIS FOR PROJECT 3

// how good is this GameState for the runner and the ghosts?
// bigger is better for the runner (up to a max of UTILITY_RUNNER_WIN), 
// smaller is better for the ghosts (down to a min of UTIILTY_GHOST_WIN)

float RunnerAgent::eval(const GameState & GS, BlockMaze *M)
{
  int i;
  Point_2D runner_position;
  int closest_pill_distance, closest_pill_index;
  int closest_ghost_distance, closest_ghost_index;
  bool is_power_pill, is_ghost;

  // BFS from runner point of view

  for (i = 0; i < GS.object_name.size(); i++) 
    if (is_runner(GS.object_name[i])) {
      runner_position = GS.object_position[i];
      break;
    }

  BFS(runner_position, M);

  // how far is nearest power pill to runner?

  is_power_pill = closest_power_pill(closest_pill_distance, closest_pill_index, GS, M);

  // how far is nearest ghost to runner?

  is_ghost = closest_ghost(closest_ghost_distance, closest_ghost_index, GS, M);

  if (is_power_pill) {
    if (is_ghost) {
      if (GS.blue_ghost_wait_counter == 0 && closest_ghost_distance < 5)  // 5
	//	return UTILITY_GHOST_WIN + closest_ghost_distance;
	return -max_path_length(GS.initial_num_power_pills, M) + closest_ghost_distance;
      //return closest_ghost_distance;
      else
	//	return UTILITY_RUNNER_WIN - closest_pill_distance;
	//return max_path_length(GS.num_power_pills() - 1, M) - closest_pill_distance;
	return max_path_length(GS.initial_num_power_pills, M) - max_path_length(GS.num_power_pills() - 1, M) - closest_pill_distance;
      //return -closest_pill_distance;
    }
    else {
      //      return UTILITY_RUNNER_WIN - closest_pill_distance;
      //      return max_path_length(GS.num_power_pills() - 1, M) - closest_pill_distance;
      return max_path_length(GS.initial_num_power_pills, M) - max_path_length(GS.num_power_pills() - 1, M) - closest_pill_distance;
    }
  }
  else
    //    return UTILITY_RUNNER_WIN;
    return max_path_length(GS.initial_num_power_pills, M);
}

//----------------------------------------------------------------------------

// OPTIONAL HELPER FUNCTION THAT I WROTE

void RunnerAgent::update_best_minimax_move(bool move_runner_turn, int move_depth, int move_agent_index, int move_delta_row, int move_delta_col, const GameState & GS, BlockMaze *M, float & best_value, Point_2D & best_move)
{
  Point_2D move_pos, temp_pos;
  GameState move_GS;
  float move_value;
  int which_pill, which_runner;

  move_pos.row = GS.object_position[move_agent_index].row + move_delta_row;
  move_pos.col = GS.object_position[move_agent_index].col + move_delta_col;

  if (!M->is_wall(move_pos.row, move_pos.col)) {

#ifdef PRINT_DEBUG
    printf("successor (%i, %i)\n", move_pos.row, move_pos.col);
#endif

    move_GS = GS;
    move_GS.object_last_position[move_agent_index] = move_GS.object_position[move_agent_index];
    move_GS.object_position[move_agent_index] = move_pos;
#ifdef PRINT_DEBUG
    move_GS.print();
#endif
    if (move_GS.check_for_power_pill_eaten(which_pill, which_runner)) {
#ifdef PRINT_DEBUG
      printf("removing power pill at (%i, %i)\n", move_GS.object_position[which_pill].row, move_GS.object_position[which_pill].col); 
#endif
      move_GS.remove_object(which_pill);
    }
    if (move_GS.blue_ghost_wait_counter > 0)
      move_GS.blue_ghost_wait_counter--;

    // what is minimax value of this "successor" state and
    // is it better than what we've seen so far?

    minimax_value(!move_runner_turn, move_depth + 1, move_GS, M, move_value, temp_pos);

    //    printf("TRYING: value %.1f\n", move_value);
    //    move_GS.print();

    // if it's runner's turn, best is MAX

    if (move_runner_turn) {
      if (move_value > best_value) {
	best_value = move_value;
	best_move = move_pos;
      }
    }

    // if it's ghost's turn, best is MIN

    else {
      if (move_value < best_value) {
	best_value = move_value;
	best_move = move_pos;
      }
    }
  }
}

//----------------------------------------------------------------------------

// STUDENTS WRITE THIS FOR PROJECT 3

// runner_turn: is this level of game tree runner making a move or ghost?
// depth: how far down the game tree are we (0 = root)
// GS, M: state of game to compute value for
// best_value: what is the computed value? (should be filled in by this function)
// best_move: where should the agent whose turn it is go? (should be filled in by this function)

// return value = true: best_move has been filled in (i.e., there is a move to make)
// return value = false: there is no best_move because GS is a leaf of game tree
 
bool RunnerAgent::minimax_value(bool runner_turn, int depth, const GameState & GS, BlockMaze *M, float & best_value, Point_2D & best_move)
{
  int i, runner_index;

  // which object is ME?
  
  for (i = 0; i < GS.object_name.size(); i++) 
    if (is_runner(GS.object_name[i])) {
      runner_index = i;
      break;
    }

  // is the game over?
  
  if (terminal_state(best_value, GS, M))   // if true, utility is filled in
    return false;

  // are we at the depth limit of our search?

  else if (cutoff_test(depth, GS, M)) {
    best_value = eval(GS, M);
#ifdef PRINT_DEBUG
    printf("EVAL = %.1f (%i, %i)\n", best_value, best_move.row, best_move.col);
#endif
    return false;
  }

  // runner's turn: which successor of GS has MAX minimax_value()?

  else if (runner_turn) {

    // return max over s in successors(GS) of minimax_value(s)

    //    best_value = UTILITY_GHOST_WIN - 1;
    best_value = -max_path_length(GS.initial_num_power_pills, M);
#ifdef PRINT_DEBUG
    printf("RUNNER NORTH\n----------------------------------\n");
#endif
    update_best_minimax_move(runner_turn, depth, runner_index, -1,  0, GS, M, best_value, best_move);
#ifdef PRINT_DEBUG
    printf("RUNNER SOUTH\n----------------------------------\n");
#endif
    update_best_minimax_move(runner_turn, depth, runner_index, +1,  0, GS, M, best_value, best_move);
#ifdef PRINT_DEBUG
    printf("RUNNER EAST\n----------------------------------\n");
#endif
    update_best_minimax_move(runner_turn, depth, runner_index,  0, +1, GS, M, best_value, best_move);
#ifdef PRINT_DEBUG
    printf("RUNNER WEST\n----------------------------------\n");
#endif
    update_best_minimax_move(runner_turn, depth, runner_index,  0, -1, GS, M, best_value, best_move);

    return true;
  }

  // ghosts' turn: which successor of GS has MIN minimax_value()?

  else {

    BFS(GS.object_position[runner_index], M);

    int closest_ghost_distance, closest_ghost_index;

    // which object is closest ghost?

    bool is_ghost = closest_ghost(closest_ghost_distance, closest_ghost_index, GS, M);

    // no ghosts -- null move

    if (!is_ghost)
      return minimax_value(!runner_turn, depth + 1, GS, M, best_value, best_move);

    // return min over s in successors(GS) of minimax_value(s)
    
    else {

      //      best_value = UTILITY_RUNNER_WIN + 1;
      best_value = max_path_length(GS.initial_num_power_pills, M);
#ifdef PRINT_DEBUG
      printf("GHOST NORTH\n----------------------------------\n");
#endif
      update_best_minimax_move(runner_turn, depth, closest_ghost_index, -1,  0, GS, M, best_value, best_move);
#ifdef PRINT_DEBUG
      printf("GHOST SOUTH\n----------------------------------\n");
#endif
      update_best_minimax_move(runner_turn, depth, closest_ghost_index, +1,  0, GS, M, best_value, best_move);
#ifdef PRINT_DEBUG
      printf("GHOST EAST\n----------------------------------\n");
#endif
      update_best_minimax_move(runner_turn, depth, closest_ghost_index,  0, +1, GS, M, best_value, best_move);
#ifdef PRINT_DEBUG
      printf("GHOST WEST\n----------------------------------\n");
#endif
      update_best_minimax_move(runner_turn, depth, closest_ghost_index,  0, -1, GS, M, best_value, best_move);

      return true;
    }
  }
}

//----------------------------------------------------------------------------

// ALL STUDENTS WRITE THIS FOR PROJECT 3

// ad-hoc ok for "Ghosts" scenario, but must use minimax for "Power Pills Work!" and "The Arena"

void RunnerAgent::update(const GameState & GS, BlockMaze *M)
{
  // uncomment to use minimax method

  if (minimax_value(true, 0, GS, M, best_minimax_value, best_minimax_move)) {
#ifdef PRINT_DEBUG
    printf("MOVING TO (%i, %i) *************************\n", best_minimax_move.row, best_minimax_move.col);
#endif
    attempt_move(best_minimax_move.row, best_minimax_move.col, M);
  }

  // uncomment for total manual control
  return;

  int closest_pill_distance, closest_pill_index;

  // we have no current plan, so make one

  if (path.size() == 0) {

    // label every vertex in maze with minimum distance from this agent
    
    BFS(position, M);
    
    // plan path to nearest power pill if it exists

    if (closest_power_pill(closest_pill_distance, closest_pill_index, GS, M))
      compute_BFS_path(GS.object_position[closest_pill_index], path, M);
    else
      return;

  }

  // get rid of null move if any

  if (path[0].x == position.x && path[0].y == position.y)
    path.erase(path.begin());    

  // try to execute next step in plan...

  if (attempt_move(path[0].y, path[0].x, M))
    path.erase(path.begin());    // erase that move from plan if successful

}

//----------------------------------------------------------------------------
//----------------------------------------------------------------------------

