//----------------------------------------------------------------------------
// pacman-like game
// copyright 2010 christopher rasmussen
// university of delaware
// v. 1.0

//----------------------------------------------------------------------------

#include "TeamRavenousZombies.hh"

//----------------------------------------------------------------------------

void get_adjacent(Point_2D start, BlockMaze *M, vector < BlockVertex* > & v){
  v.clear();
  BlockVertex* w;
  w = &(M->vertex[start.row][start.col]);

  if(start.row+1 < M->num_rows){
    w = &(M->vertex[start.row+1][start.col]);
    if(w->is_empty())
      v.push_back(w);
  }

  if(start.col+1 < M->num_cols){
    w = &(M->vertex[start.row][start.col+1]);
    if(w->is_empty())
      v.push_back(w);
  }

  if(start.row-1 > -1){
    w = &(M->vertex[start.row-1][start.col]);
    if(w->is_empty())
      v.push_back(w);
  }

  if(start.col-1 > -1){
    w = &(M->vertex[start.row][start.col-1]);
    if(w->is_empty())
      v.push_back(w);
  }

}

void get_adjacent(Point_2D start, BlockMaze *M, vector < Point_2D > & v){
  v.clear();
  BlockVertex* w;
  w = &(M->vertex[start.row][start.col]);

  if(start.row+1 < M->num_rows){
    w = &(M->vertex[start.row+1][start.col]);
    if(w->is_empty())
      v.push_back( Point_2D(start.row+1,start.col) );
  }

  if(start.col+1 < M->num_cols){
    w = &(M->vertex[start.row][start.col+1]);
    if(w->is_empty())
      v.push_back( Point_2D(start.row,start.col+1) );
  }

  if(start.row-1 > -1){
    w = &(M->vertex[start.row-1][start.col]);
    if(w->is_empty())
      v.push_back( Point_2D(start.row-1,start.col) );
  }

  if(start.col-1 > -1){
    w = &(M->vertex[start.row][start.col-1]);
    if(w->is_empty())
      v.push_back( Point_2D(start.row,start.col-1) );
  }

}

bool odd_enemy(Point_2D p, const GameState & GS){
  bool enemy = false;
  for (int i = 0; i < GS.object_position.size(); i++){
    if(p.row == GS.object_position[i].row
       && p.col == GS.object_position[i].col
       && is_odd_team(GS.object_name[i]) && p.col < 16)
      enemy = true;
  }
  return enemy;

}


bool even_enemy(Point_2D p, const GameState & GS){
  bool enemy = false;
  for (int i = 0; i < GS.object_position.size(); i++){
    if(p.row == GS.object_position[i].row
       && p.col == GS.object_position[i].col
       && is_even_team(GS.object_name[i]) && p.col > 16)
      enemy = true;
  }
  return enemy;
}



bool enemy(Point_2D p, const GameState & GS, string name){
  
  if(is_even_team(name))
    return odd_enemy(p,GS);
  else if(is_odd_team(name))
    return even_enemy(p,GS);
  else
    return false;
}

bool is_adjacent(Point_2D p, const GameState & GS, string name){
  if(enemy(Point_2D(p.row+1,p.col),GS,name))
    return true;
  else if(enemy(Point_2D(p.row-1,p.col),GS,name))
    return true;
  else if(enemy(Point_2D(p.row,p.col+1),GS,name))
    return true;
  else if(enemy(Point_2D(p.row,p.col-1),GS,name))
    return true;
  else
    return false;
}

bool adjacent_to_adjacent(Point_2D p, const GameState & GS, string name){
  if(is_adjacent(Point_2D(p.row+1,p.col),GS,name))
    return true;
  else if(is_adjacent(Point_2D(p.row-1,p.col),GS,name))
    return true;
  else if(is_adjacent(Point_2D(p.row,p.col+1),GS,name))
    return true;
  else if(is_adjacent(Point_2D(p.row,p.col-1),GS,name))
    return true;
  else
    return false;
}

bool friendly(Point_2D p, const GameState & GS,string name){
  bool f = false;
  for (int i = 0; i < GS.object_position.size(); i++){
    if(p.row == GS.object_position[i].row
       && p.col == GS.object_position[i].col
       && (is_even_team(GS.object_name[i])==is_even_team(name)))
      f = true;
  }
  return f;
}


//computes the cost of moving FROM a INTO b
int cost(BlockVertex* a, BlockVertex* b, const GameState & GS, string name){

  if(enemy(a->pos,GS,name))
    return 40;
  else if(is_adjacent(a->pos,GS,name))
    return 28;
  else if(friendly(a->pos,GS,name))
    return 10;
  else if(a->is_fire())
    return FIRE_COST;
  else if(adjacent_to_adjacent(a->pos,GS,name))
    return 6;
  else if(a->is_water())
    return WATER_COST;
  else
    return 1;
}


class BV_Compare{
public:
  bool operator()(BlockVertex* a, BlockVertex* b){
    if(b->distance == DIST_INFINITY)
      return false;
    else if( a->distance == DIST_INFINITY )
      return true;
    else
      return a->distance > b->distance;
      
  }
};

void Dijkstra(Point_2D start, BlockMaze *M, const GameState & GS, string name)
{
  //M->initialize_distances();
  
  for(int r(0); r<M->num_rows; r++){
    for(int c(0); c<M->num_cols; c++){
      M->vertex[r][c].known = false;
      M->vertex[r][c].distance = DIST_INFINITY;
      M->vertex[r][c].path = NULL;
    }
  }
  
  BlockVertex* v;
  vector < BlockVertex* > adjacent;
  Point_2D curr_pos;
  BlockVertex* start_v = &(M->vertex[start.row][start.col]);

  priority_queue< BlockVertex*, vector<BlockVertex*>, BV_Compare > pq;

  
  
  start_v->distance = 0;
  pq.push(start_v);

  while(!pq.empty()){
    v = pq.top();
    pq.pop();
    if(!v->known){
      v->known = true;
      get_adjacent(v->pos, M, adjacent);
      for(int i(0); i<adjacent.size(); i++){
	if(!adjacent[i]->known){
	  if(adjacent[i]->distance == DIST_INFINITY || adjacent[i]->distance > v->distance + cost(v,adjacent[i],GS,name) ){
	    adjacent[i]->distance = v->distance + cost(v,adjacent[i],GS,name);
	    adjacent[i]->path = v;
	    pq.push(adjacent[i]);
	  }
	}
      }
    }

  }

}

//----------------------------------------------------------------------------

// ALL STUDENTS WRITE THIS FOR PROJECT 3

// ad-hoc or minimax ok for "The Arena"

void TeamRavenousZombies::update(const GameState & GS, BlockMaze *M)
{

  Dijkstra(position,M,GS,name);


//   bool layered = false;
//   for(int k = 0; k < GS.object_name.size(); k++){
//     if(GS.object_name[k].find("runner") != string::npos &&
//        position.row == GS.object_position[k].row &&
//        position.col == GS.object_position[k].col &&
//        (is_odd_team(name) == is_odd_team(GS.object_name[k])) &&
//        name != GS.object_name[k])
//       layered = true;
//   }

//   bool was_layered = false;
//   for(int k = 0; k < GS.object_name.size(); k++){
//     if(GS.object_name[k].find("runner") != string::npos &&
//        last_position.row == GS.object_last_position[k].row &&
//        last_position.col == GS.object_last_position[k].col &&
//        (is_odd_team(name) == is_odd_team(GS.object_name[k])) &&
//        name != GS.object_name[k])
//       was_layered = true;
//   }

  

//   if(layered || (last_position.row == position.row && last_position.col == position.col)){
//     int t = rand()%GS.object_position.size();
//     cout << t;
//     target = GS.object_position[t];
//   }

//   if(layered || was_layered || (last_position.row == position.row && last_position.col == position.col)){
//     attempt_move(target.row,target.col,M);
//     return;
//   }

  int i,j,l,m;  
  Point_2D p;

  if(is_odd_team(name)){
    closest_even_runner(i,j,GS,M);
    closest_odd_power_pill(l,m,GS,M);
    

    if( i < l  && GS.object_position[j].col < 16){
      compute_BFS_path(GS.object_position[j],path,M);
    }
    else
      compute_BFS_path(GS.object_position[m],path,M);
  }  
  if(is_even_team(name)){
    closest_odd_runner(i,j,GS,M);
    closest_even_power_pill(l,m,GS,M);
    
    if( i < l  && GS.object_position[j].col > 16){
      compute_BFS_path(GS.object_position[j],path,M);
    }
    else
      compute_BFS_path(GS.object_position[m],path,M);
  }  

  attempt_move(path[1].row,path[1].col,M);
}

//----------------------------------------------------------------------------
//----------------------------------------------------------------------------

