//----------------------------------------------------------------------------
// pacman-like game
// copyright 2010 christopher rasmussen
// university of delaware
// v. 1.0

//----------------------------------------------------------------------------

#include "TeamTardis.hh"
#include <fstream>
#include <cstdlib>
#include <stdio.h>
#include <stdlib.h>

//----------------------------------------------------------------------------
//----------------------------------------------------------------------------

const int MAXIMUM_DEPTH_FOR_MINIMAX = 1;

//----------------------------------------------------------------------------

string itoa(int n){
  
  char * s = new char[17];
  string u;
  
  if (n < 0) {
    n = (-1 * n);
    u = "-"; 
  }
  
  int i = 0;
  
  do {
    s[i++]= n % 10 + '0';
    n -= n % 10;
  } while ((n /= 10) > 0);
  
  for (int j = i - 1; j >= 0; j--)
    u += s[j];
  
  delete[] s;
  return u;
}

//----------------------------------------------------------------------------

// ALL STUDENTS WRITE THIS FOR PROJECT 3

// ad-hoc or minimax ok for "The Arena"

void TeamTardis::update(const GameState & GS, BlockMaze *M)
{
  // uncomment for total manual control
  //  return;

  float best_minimax_value;
  Point_2D best_minimax_move = position;

  if(minimax_value(true, 0, GS, M, best_minimax_value, best_minimax_move))
    attempt_move(best_minimax_move.row, best_minimax_move.col, M);

}

//----------------------------------------------------------------------------

bool minimax_value(bool runner_turn, int depth, const GameState & GS, BlockMaze *M, float & best_value, Point_2D best_move)
{

  // runner_turn == true denotes the odd team's turn.
  // runner_turn == false denotes the even team's turn.
  
  Point_2D runnerPosition;

  if(terminal_state(best_value, GS, M))
     return false;


  else if (cutoff_test(depth, GS, M)){
    best_value = eval(GS, M);
    return false;
  }

  else if(runner_turn){
    runnerPosition = best_move;
    // odd team's turn 
    depth++;

    vector <Point_2D> oddRunners;
    int agentID;
    int whichAgent;
    for(int i = 0; i < GS.object_name.size(); i++){
      if(GS.object_name[i].find("odd") != string::npos){
	if (best_move.row != GS.object_position[i].row && best_move.col != GS.object_position[i].col){
	  oddRunners.push_back(GS.object_position[i]);
	}
	else{
	  agentID = atoi(&GS.object_name[i][7]);
	  whichAgent = i;
	}
      }
    }
    int nextAgentID = agentID + 1;
    string buffer = itoa(nextAgentID);
    
    //find the nextAgentID
    Point_2D nextAgentPosition;
    
    for (int k = 0; k < GS.object_name.size(); k++){
      if (GS.object_name[k].find(buffer[1]) != string::npos){
	nextAgentPosition = GS.object_position[k];
      }
    }
    // see which of the four moves that are possible gives you the best value
    priority_queue <float> pq;
    GameState moveN = GS;
    float northBV;
    Point_2D newNP;
    bool Npossible = false;
    if(create_game_state('N', moveN, M, newNP, true, whichAgent)){
      Npossible = true;
      minimax_value(0, depth, moveN, M, northBV, nextAgentPosition);
      pq.push(northBV);
    }
    
    GameState moveS = GS;
    float southBV;
    Point_2D newSP;
    bool Spossible = false;
    if (create_game_state('S', moveS, M, newSP, true, whichAgent)){
      Spossible = true;
      minimax_value(0, depth, moveS, M, southBV, nextAgentPosition);
      pq.push(southBV);
    }
    
    
    GameState moveE = GS;
    float eastBV;
    Point_2D newEP;
    bool Epossible = false;
    if (create_game_state('E', moveE, M, newEP, true, whichAgent)){
      Epossible = true;
      minimax_value(0, depth, moveE, M, eastBV, nextAgentPosition);
      pq.push(eastBV);
    }
    
    GameState moveW = GS;
    float westBV;
    Point_2D newWP;
    bool Wpossible = false;
    if (create_game_state('W', moveW, M, newWP, true, whichAgent)){
      Wpossible = true;
      minimax_value(0, depth, moveW, M, westBV, nextAgentPosition);
      pq.push(westBV);
    }

    best_value = pq.top();
    
    if (depth == 1){
      if (best_value == northBV && Npossible)
	best_move = newNP;
      else if(best_value == southBV && Spossible)
	best_move = newSP;
      else if (best_value == westBV && Wpossible)
	best_move = newWP;
      else 
	best_move = newEP;
    }
   
    return true;

  }

  else{
    int agentID;
    // even team's turn
    runnerPosition = best_move;
    depth++;
    int whichAgent;
    vector <Point_2D> evenRunners;
    for(int i = 0; i < GS.object_name.size(); i++){
      if(GS.object_name[i].find("even") != string::npos){
	if (GS.object_position[i].row != runnerPosition.row && GS.object_position[i].col != runnerPosition.col)
	  evenRunners.push_back(GS.object_position[i]);
	else{
	  agentID = atoi(&GS.object_name[i][7]);
	  whichAgent = i;
	}
      }
    }
    int nextAgentID = agentID + 1;
    string buffer = itoa(nextAgentID);
    
    //find the nextAgentID
    Point_2D nextAgentPosition;
    
    for (int k = 0; k < GS.object_name.size(); k++){
      if (GS.object_name[k].find(buffer[1]) != string::npos){
	nextAgentPosition = GS.object_position[k];
      }
    }
    

    //find move one with lowest value 
    float bestValueSoFar = 10000000;
    GameState moveN = GS;
    float northBV;
    Point_2D newNP;
    bool Npossible = false;
    if(create_game_state('N', moveN, M, newNP, false, whichAgent)){
      Npossible = true;
      minimax_value(0, depth, moveN, M, northBV, nextAgentPosition);
      if (northBV < bestValueSoFar)
	bestValueSoFar = northBV;
    }
    
    GameState moveS = GS;
    float southBV;
    Point_2D newSP;
    bool Spossible = false;
    if (create_game_state('S', moveS, M, newSP, false, whichAgent)){
      Spossible = true;
      minimax_value(0, depth, moveS, M, southBV, nextAgentPosition);
      if (southBV < bestValueSoFar)
	bestValueSoFar = southBV;
    }
    
    
    GameState moveE = GS;
    float eastBV;
    Point_2D newEP;
    bool Epossible = false;
    if (create_game_state('E', moveE, M, newEP, false, whichAgent)){
      Epossible = true;
      minimax_value(0, depth, moveE, M, eastBV, nextAgentPosition);
      if (eastBV < bestValueSoFar)
	bestValueSoFar = eastBV;
    }
    
    GameState moveW = GS;
    float westBV;
    Point_2D newWP;
    bool Wpossible = false;
    if (create_game_state('W', moveW, M, newWP, false, whichAgent)){
      Wpossible = true;
      minimax_value(0, depth, moveW, M, westBV, nextAgentPosition);
      if (westBV < bestValueSoFar)
	bestValueSoFar = westBV;
    }
    
    best_value = bestValueSoFar;

    if (depth == 1){
      if (best_value == northBV && Npossible)
	best_move = newNP;
      else if(best_value == southBV && Spossible)
	best_move = newSP;
      else if (best_value == westBV && Wpossible)
	best_move = newWP;
      else 
	best_move = newEP;
    }

    ofstream OutStream;
    OutStream.open("debug.txt");
    OutStream << best_value << endl;
    OutStream.close();
   
    return true;
  }
}

  
//----------------------------------------------------------------------------

float eval(const GameState & GS, BlockMaze *M)
{

  // teams
  vector <Point_2D> evenOffence;
  vector <Point_2D> evenDefence;
  vector <Point_2D> oddOffence;
  vector <Point_2D> oddDefence;

  int agentID;
  bool isDefence;

  for(int i = 0; i < GS.object_name.size(); i++){

    agentID = atoi(&GS.object_name[i][7]);

    // maps a natural number to a 0 or 1
    // 0 = 0, 1 = 0, 2 = 1, 3 = 1, 4 = 0, 5 = 0, ...
    isDefence = ((agentID / 2) % 2);

    // decide what team and role this runner has
    if(GS.object_name[i].find("even") != string::npos)
      if(isDefence) evenDefence.push_back(GS.object_position[i]);
      else evenOffence.push_back(GS.object_position[i]);
    else if (GS.object_name[i].find("odd") != string::npos)
      if(isDefence) oddDefence.push_back(GS.object_position[i]);
      else oddOffence.push_back(GS.object_position[i]);

  }
  
  
  // runners in enemy territory
  vector <Point_2D> closeAvailEvenRunners;
  vector <Point_2D> closeAvailOddRunners;
  
  // pills to eat
  vector <Point_2D> closeEvenPills;
  vector <Point_2D> closeOddPills;
  
  // defenders
  vector <Point_2D> closeEvenRunners;
  vector <Point_2D> closeOddRunners;

  //generate list of runners on their own sides & pills on their respective sides
  for (int k = 0; k < GS.object_name.size(); k++){
    if(GS.object_name[k].find("even") != string::npos)
      if(M->is_even_side(GS.object_position[k]))
	closeEvenRunners.push_back(GS.object_position[k]);
      else
	closeAvailEvenRunners.push_back(GS.object_position[k]);
    else if(GS.object_name[k].find("odd") != string::npos)
      if(M->is_odd_side(GS.object_position[k]))
	closeOddRunners.push_back(GS.object_position[k]);   
      else
	closeAvailOddRunners.push_back(GS.object_position[k]);
    else if (GS.object_name[k].find("power pill") != string::npos)
      if(M->is_even_side(GS.object_position[k]))
	closeOddPills.push_back(GS.object_position[k]);
      else
	closeEvenPills.push_back(GS.object_position[k]);
  }

  //EVEN

  //offence
  int distanceToClosestThreatE;
  int distanceToClosestPillE = 0;
  int distanceToClosestPilll;
  //what's the closest agent that could eat the agents && what's the closest power pill
  for (int r = 0; r < closeAvailEvenRunners.size(); r++){
    Dijkstra(closeAvailEvenRunners[r], M);
    distanceToClosestPilll = M->num_rows*M->num_cols;
    for (int d = 0; d < closeOddRunners.size(); d++){
      if (closest_odd_runner(closeOddRunners[d].row, closeOddRunners[d].col, GS, M)){
	distanceToClosestThreatE += M->vertex[closeOddRunners[d].row][closeOddRunners[d].col].distance;
      }
    }
    for (int e = 0; e < closeEvenPills.size(); e++){
      if (M->vertex[closeEvenPills[e].row][closeEvenPills[e].col].distance < distanceToClosestPilll){
	distanceToClosestPilll = M->vertex[closeEvenPills[e].row][closeEvenPills[e].col].distance;
      }
    }
    distanceToClosestPillE += distanceToClosestPilll;
  }
  //defence
  //what's the closest agent to omnomnom
  int closestNomE = 0;
  int closestNomm;
  for (int s = 0; s < closeEvenRunners.size(); s++){
    Dijkstra(closeEvenRunners[s], M);
    closestNomm = M->num_rows*M->num_cols;
    for (int w = 0; w < closeAvailOddRunners.size(); w++){
      if(M->vertex[closeAvailOddRunners[w].row][closeAvailOddRunners[w].col].distance < closestNomm){
	closestNomm = M->vertex[closeAvailOddRunners[w].row][closeAvailOddRunners[w].col].distance;
      }
    }
    closestNomE += closestNomm;
  }

  //ODD

  //offence
  int distanceToClosestThreatO = 0;
  int distanceToClosestPillO = 0;
  int distLL;
  //what's the closest agent that could eat the agents && what's the closest power pill
  for (int y = 0; y < closeAvailOddRunners.size(); y++){
    Dijkstra(closeAvailOddRunners[y], M);
    for (int g = 0; g < closeEvenRunners.size(); g++){
      distLL = M->num_rows*M->num_cols;
      if (closest_even_runner(closeEvenRunners[g].row, closeEvenRunners[g].col, GS, M)){
	distanceToClosestThreatO += M->vertex[closeEvenRunners[g].row][closeEvenRunners[g].col].distance;
      }
    }
    for (int h = 0; h < closeOddPills.size(); h++){
      if (M->vertex[closeOddPills[h].row][closeOddPills[h].col].distance < distLL){
	distLL = M->vertex[closeOddPills[h].row][closeOddPills[h].col].distance;
      }
    }
  }

  //defence
  int closestNomO = 0;
  int dontCare;
  for (int f = 0; f < closeOddRunners.size(); f++){
    Dijkstra(closeOddRunners[f], M);
    for (int q = 0; q < closeAvailEvenRunners.size(); q++){
      if(dontCare > M->vertex[closeAvailEvenRunners[q].row][closeAvailEvenRunners[q].col].distance){
	dontCare = M->vertex[closeAvailEvenRunners[q].row][closeAvailEvenRunners[q].col].distance;
      }
    }
    closestNomO += dontCare;
  }
  float value;

  //min = odd, max = even

  value = -distanceToClosestPillE + distanceToClosestPillO;
  

}

//----------------------------------------------------------------------------

bool terminal_state(float & utility, const GameState & GS, BlockMaze *M)
{

  if(GS.arena_check_for_odd_level_cleared(M)){
    utility = -100; // change with UTILITY_EVEN_WIN once you figure that out
    return true;
  }
  else if(GS.arena_check_for_even_level_cleared(M)){
    utility = 100; // change with UTILITY_ODD_WIN once you figure that out
    return true;
  }
  else 
    return false;

}

//----------------------------------------------------------------------------

bool cutoff_test(int depth, const GameState & GS, BlockMaze *M)
{

  if (depth >= MAXIMUM_DEPTH_FOR_MINIMAX)
    return true;
  else
    return false;

}

//----------------------------------------------------------------------------


bool create_game_state(char dir, GameState & GS, BlockMaze* M, Point_2D & newPos, bool odd, int index){
  bool isValidMove = true; 
  Point_2D prevPos;
  Point_2D curPos;

  curPos = GS.object_position[index];
  prevPos = GS.object_last_position[index];
  GS.object_last_position[index] = curPos;

  if (dir == 'N'){
    curPos.row++;
    GS.object_position[index] = curPos;
  }
  else if (dir == 'S'){
    curPos.row--;
    GS.object_position[index] = curPos;
  }
  else if (dir == 'E'){
    curPos.col++;
    GS.object_position[index] = curPos;
  }
  else{
    curPos.col--;
    GS.object_position[index] = curPos;
  }

  if (M->vertex[curPos.row][curPos.col].is_wall())
    isValidMove = false;
  else if (curPos.row == prevPos.row && curPos.col == prevPos.col)
    isValidMove = false;

  
  return isValidMove;
}

//------------------------------------------------------------

int cost(BlockVertex next){
  int cost;
  if (next.is_none())
    cost = 1;
  else if (next.is_fire())
    cost = FIRE_COST;
  else if (next.is_water())
    cost = WATER_COST;
  return cost;
}
